// -*- C++ -*-
#include "Rivet/Analysis.hh"
#include "Rivet/Projections/FinalState.hh"
#include "Rivet/Projections/IdentifiedFinalState.hh"
#include "Rivet/Projections/PromptFinalState.hh"
#include "Rivet/Projections/VetoedFinalState.hh"
#include "Rivet/Projections/DressedLeptons.hh"
#include "Rivet/Projections/FastJets.hh"


namespace Rivet {

  class ATLAS_2016_I1449082 : public Analysis {
  public:

    enum MeasurType {
      kInclMeas,
      kmttMeas,
      kbetaMeas,
      kptMeas,
      kNmeas
    };
    
    const double static MW ;
    const double static Mtop ;

    const int static kNbins = 2;

    const double static bins[kNmeas][kNbins+1];

    //names of the measurements and bin ranges
    const static string measStr [kNmeas] ; 
    const static string rangeStr[kNmeas][kNbins];

    /// Constructor
    DEFAULT_RIVET_ANALYSIS_CTOR(ATLAS_2016_I1449082);

    /// @name Analysis methods
    //@{

    /// Book histograms and initialise projections before the run
    void init() {

      /// Initialise and register projections here

      // ---------------------
      // electron projections
      // ---------------------
      //Electron/muons are defined from electron/muon and photons within a cone of ∆R = 0.1. 
      //No isolation condition is imposed. The parent of the electron/muon is required not to be a hadron or quark. 

      // Eta ranges
      Cut eta_full = (Cuts::abseta < 5.0);
      //only in rel2.4
      //Cut lep_cuts = (Cuts::abseta < 2.5) && (Cuts::pT > 25*GeV);
      Cut lep_cuts = (Cuts::abseta < 2.5) & (Cuts::pT > 25*GeV);

      // All final state particles
      FinalState fs(eta_full);

      // Get photons to dress leptons
      IdentifiedFinalState photons(fs);
      photons.acceptIdPair(PID::PHOTON);

      // Projection to find the electrons
      IdentifiedFinalState el_id(fs);
      el_id.acceptIdPair(PID::ELECTRON);

      PromptFinalState electrons(el_id);
      electrons.acceptTauDecays(true);

      DressedLeptons dressedelectrons(photons, electrons, 0.1, lep_cuts, true, true);
      declare(dressedelectrons, "dressedelectrons");

      DressedLeptons dressedelectrons_full(photons, electrons, 0.1, eta_full, true, true);

      // ---------------------
      // muons projections
      // ---------------------
      IdentifiedFinalState mu_id(fs);
      mu_id.acceptIdPair(PID::MUON);

      PromptFinalState muons(mu_id);
      muons.acceptTauDecays(true);
      
      DressedLeptons dressedmuons(photons, muons, 0.1, lep_cuts, true, true);
      declare(dressedmuons, "dressedmuons");

      DressedLeptons dressedmuons_full(photons, muons, 0.1, eta_full, true, true);

      // ---------------------
      // neutrinos  projections
      // ---------------------
      //Missing ET is calculated as the 4–vector sum of neutrinos from W/Z-boson decays. Tau decays are
      //included. A neutrino is treated as a detectable particle and is selected for consideration in the same
      //way as electrons or muons, i.e. the parent is required not to be a hadron or quark (u − b).
      IdentifiedFinalState nu_id;
      nu_id.acceptNeutrinos();

      PromptFinalState neutrinos(nu_id);
      neutrinos.acceptTauDecays(true); 
      declare(neutrinos, "neutrinos");

      // ---------------------
      // jets projections
      // ---------------------
      //Jets are defined with the anti–kt algorithm, clustering all stable particles excluding the electrons,
      //muons, neutrinos, and photons used in the definition of the selected leptons.
      VetoedFinalState vfs(fs);
      vfs.addVetoOnThisFinalState(dressedelectrons_full);
      vfs.addVetoOnThisFinalState(dressedmuons_full);
      vfs.addVetoOnThisFinalState(neutrinos);
      addProjection( FastJets(vfs, FastJets::ANTIKT, 0.4), "Jets");

      //-------------------------
      /// Book histograms here
      //-------------------------
      _h_dEta      = bookHisto1D(1, 1, 1);
      _h_dY        = bookHisto1D(2, 1, 1);

      for (int iM = 0; iM < kNmeas; ++iM) {
        _h_Acll[iM] = bookScatter2D(3+iM, 1, 1);
        _h_Actt[iM] = bookScatter2D(7+iM, 1, 1);
      }

      for (int iM = 0; iM < kNmeas; ++iM) {
        for (int iB = 0; iB < kNbins; ++iB) {
          _h_dEta_asym[iM][iB] = bookHisto1D( "dEta_asym_" + measStr[iM] + "_bin" + rangeStr[iM][iB],  2,  -10., 10.);
          _h_dY_asym  [iM][iB] = bookHisto1D( "dY_asym_"   + measStr[iM] + "_bin" + rangeStr[iM][iB],  2,  -10., 10.);
        }
      }
    }


    /// Perform the per-event analysis
    void analyze(const Event& event) {

      const double weight = event.weight();

      //---------------------------------------
      //get electrons/muons and apply selection
      //---------------------------------------

      // Get the selected objects, using the projections.
      vector<DressedLepton> dressedelectrons = apply<DressedLeptons>(event, "dressedelectrons").dressedLeptons();
      vector<DressedLepton> dressedmuons     = apply<DressedLeptons>(event, "dressedmuons").dressedLeptons();

      vector<DressedLepton>  leptons;
      leptons.reserve( dressedelectrons.size() + dressedmuons.size() ); 
      leptons.insert( leptons.end(), dressedelectrons.begin(), dressedelectrons.end() );
      leptons.insert( leptons.end(), dressedmuons.begin()    , dressedmuons.end() );

      // Require at least 2 leptons in the event
      if (leptons.size() < 2)  vetoEvent;
      
      //---------------------------------------
      //get neutrinos
      //---------------------------------------
      Particles neutrinos  = apply<PromptFinalState>(event, "neutrinos").particlesByPt();
      
      // Require at least 2 neutrinos in the event
      if (neutrinos.size() < 2)  vetoEvent;

      //---------------------------------------
      //get jets and apply selection
      //---------------------------------------
      const Jets alljets = apply<FastJets>(event, "Jets").jetsByPt(25*GeV);

      Jets jets;
      foreach (const Jet& j, alljets) {
        if ( j.abseta() < 2.5 ) jets.push_back(j);
      }
      
      // Require at least 2 jets in the event
      if (jets.size() < 2)  vetoEvent;
 
      //---------------------------------------
      //remaining selections
      //---------------------------------------

      //Events where leptons and jets overlap, within dR = 0.4, are rejected.
      foreach (const DressedLepton& lepton, leptons) {
        foreach (const Jet& j, jets) {
          if ( deltaR(j, lepton) <  0.4)  vetoEvent;
        }
      }

      //---------------------------------------
      //construct pseudo-tops
      //---------------------------------------

      //Exactly 2 opposite sign leptons are required (e/mu)
      if (leptons.size() != 2)  vetoEvent;

      if ( (leptons[0].charge() * leptons[1].charge()) > 0.)  vetoEvent;
      
      FourMomentum lep_p, lep_n;
      foreach (const DressedLepton& lepton, leptons) {
        if (lepton.charge() > 0.)  lep_p = lepton.momentum();
        if (lepton.charge() < 0.)  lep_n = lepton.momentum();
      }

      //Only the 2 leading pT selected neutrinos are considered 
      FourMomentum nu1 = neutrinos[0].momentum();
      FourMomentum nu2 = neutrinos[1].momentum();

      //two jets correspond to the two leading jets in the event. 
      //If there is any b-tagged jet in the event, then the b-tagged jets 
      //are preferentially selected over the non-tagged jets without taking into account its pT.
      //A jet is a b–jet if any B–hadron is included in the jet. 
      //Only B-hadrons with an initial pT > 5 GeV are considered.
      Jets bjets, lightjets;
      foreach (const Jet& jet, jets) {
        // Count the number of b-tags
        size_t b_tagged = jet.bTags(Cuts::pT > 5*GeV).size();
        if ( b_tagged )  bjets     += jet;
        else             lightjets += jet;
      }

      std::sort(bjets.begin()    , bjets.end()    , cmpMomByPt);
      std::sort(lightjets.begin(), lightjets.end(), cmpMomByPt);

      //initially take 2 highest Et jets
      FourMomentum bjet1 = jets[0].momentum();
      FourMomentum bjet2 = jets[1].momentum();
      
      if (bjets.size()) {
        bjet1 = bjets[0].momentum();
        if (bjets.size() > 1) bjet2 = bjets[1].momentum();
        else                  bjet2 = lightjets[0].momentum(); //we should have light jet because >=2 jets requirement
      } 
      else {
        //no btagged jets --> should have >= 2 light jets
        bjet1 = lightjets[0].momentum();
        bjet2 = lightjets[1].momentum();
      }
      
      //construct pseudo-W bosons from lepton-neutrino combinations
      //minimize the difference between the mass computed from each lepton-neutrino combination and the W boson mass 
      double massDiff1 = fabs( (nu1 + lep_p).mass() - MW ) + fabs( (nu2 + lep_n).mass() - MW );
      double massDiff2 = fabs( (nu1 + lep_n).mass() - MW ) + fabs( (nu2 + lep_p).mass() - MW );

      FourMomentum Wp, Wn;
      if (massDiff1 < massDiff2) {
        Wp = nu1+lep_p;
        Wn = nu2+lep_n;
      }
      else {
        Wp = nu2+lep_p;
        Wn = nu1+lep_n;
      }

      //construct pseudo-tops from jets and pseudo-W bosons
      //minimize the difference between the mass computed from each W-boson and b-jet combination and the top mass 
      massDiff1 = fabs( (Wp+bjet1).mass()*GeV - Mtop ) + fabs( (Wn+bjet2).mass()*GeV - Mtop );
      massDiff2 = fabs( (Wp+bjet2).mass()*GeV - Mtop ) + fabs( (Wn+bjet1).mass()*GeV - Mtop );
      
      FourMomentum top_p, top_n;
      if (massDiff1 < massDiff2) {
        top_p = Wp+bjet1;
        top_n = Wn+bjet2;
      }
      else {
        top_p = Wp+bjet2;
        top_n = Wn+bjet1;
      }

      //---------------------------------------
      //calculate d|eta|, d|y|, etc.
      //---------------------------------------
      double dEta = lep_p.abseta() - lep_n.abseta();
      double dY   = top_p.absrapidity() - top_n.absrapidity();

      double mtt  = (top_p + top_n).mass()*GeV;
      double beta = fabs( (top_p + top_n).pz() ) / (top_p + top_n).E();
      double pttt = (top_p + top_n).pt()*GeV;

      //---------------------------------------
      //fill histos, counters
      //---------------------------------------
      _h_dEta->fill(dEta, weight);
      _h_dY  ->fill(dY  , weight);

      //histos for inclusive and differential asymmetries
      int mttBinID  = getBinID(kmttMeas , mtt);
      int betaBinID = getBinID(kbetaMeas, beta);
      int ptttBinID = getBinID(kptMeas  , pttt);

      for (int iM=0; iM < kNmeas; ++iM) {

        int binID = -1;
        switch (iM)  {
          case kInclMeas : binID = 0;         break;
          case kmttMeas  : binID = mttBinID ; break;
          case kbetaMeas : binID = betaBinID; break;
          case kptMeas   : binID = ptttBinID; break;
          default: binID = -1; break;
        }
        if (binID >= 0) {
          _h_dY_asym  [iM][binID] ->fill(dY  , weight);
          _h_dEta_asym[iM][binID] ->fill(dEta, weight);
        }
      }
    }


    /// Normalise histograms etc., after the run
    void finalize() {

      //calculate charge asymmetries and fill them to hists
      //just for cross-check, calculate asymmetries from original dEta/dY histos
      double asym=0., err=0.;

      calcAsymAndError(_h_dEta, asym, err);
      MSG_INFO("lepton inclusive asymmetry from histo:  = " << asym << " +- " << err );

      calcAsymAndError(_h_dY, asym, err);
      MSG_INFO("ttbar inclusive asymmetry from histo:  = "  << asym << " +- " << err );

      // dEta/dY distributions: normalize to unity
      normalize(_h_dEta); 
      normalize(_h_dY); 

      for (int iM=0; iM < kNmeas; ++iM) {
        for (int iB=0; iB < kNbins; ++iB) {

          removeAnalysisObject(_h_dEta_asym[iM][iB]);
          removeAnalysisObject(_h_dY_asym[iM][iB]);

          //only one bin for inclusive measurement
          if ( (iM == kInclMeas) && (iB != 0)) continue;

          calcAsymAndError(_h_dEta_asym[iM][iB], asym, err);
          _h_Acll[iM]->addPoint( (bins[iM][iB+1] + bins[iM][iB])/2., asym, (bins[iM][iB+1] - bins[iM][iB])/2., err);

          calcAsymAndError(_h_dY_asym[iM][iB], asym, err);
          _h_Actt[iM]->addPoint( (bins[iM][iB+1] + bins[iM][iB])/2., asym, (bins[iM][iB+1] - bins[iM][iB])/2., err);
        }
      }
    }

    //@}


  private:

    void calcAsymAndError(Histo1DPtr hist, double& asym, double& err)  {

      int nBins = hist->numBins(); 

      if (nBins % 2 != 0) {
      	asym = -999; err = -999.; return;
      }
      
      double Nneg  = 0.;
      double Npos  = 0.;
      double dNneg = 0.;
      double dNpos = 0.;

      for (int iB=0; iB < nBins; ++iB) {
        if (iB < nBins/2) {
          Nneg  += hist->bin(iB).sumW();
          dNneg += hist->bin(iB).sumW2();
        } else {
          Npos  += hist->bin(iB).sumW();	  
          dNpos += hist->bin(iB).sumW2();
        }
      }
      dNneg = sqrt(dNneg);
      dNpos = sqrt(dNpos);

      asym = (Npos + Nneg) != 0.0 ? (Npos - Nneg) / (Npos + Nneg) : -999.;

      double Ntot = Npos + Nneg;
      double Ntot2 = Ntot * Ntot;
      double dNpos2 = dNpos * dNpos;
      double dNneg2 = dNneg * dNneg;

      err = Ntot2 != 0. ? 2. * sqrt( (dNneg2 * Npos * Npos + dNpos2 * Nneg * Nneg) / (Ntot2 * Ntot2)) : -999.;
    }

    int getBinID (MeasurType type, double value)  {

      for (int iBin = 0; iBin < kNbins; ++iBin)  { 
        if (value <= bins[type][iBin+1]) return iBin;
      }
      
      return -1;
    }

    /// @name Histograms
    //@{
    Histo1DPtr _h_dEta;
    Histo1DPtr _h_dY;
    Scatter2DPtr _h_Actt[kNmeas];
    Scatter2DPtr _h_Acll[kNmeas];

    //histograms to calculate the asymmetries from
    Histo1DPtr _h_dEta_asym[kNmeas][kNbins];
    Histo1DPtr _h_dY_asym  [kNmeas][kNbins];

    //@}
    
    //add some temporary histograms

    //not_scaled histos
    Histo1DPtr _h_dEta_notscaled;
    Histo1DPtr _h_dY_notscaled;

  };
  
  const double ATLAS_2016_I1449082::bins[][kNbins+1] = { { 0., 1., 2. }, //inclusive (dummy)
							 { 0., 500., 2000.}, //mtt [GeV]
							 { 0., 0.6 , 1.0}, //beta
							 { 0., 30. , 1000.} }; //pttt [GeV]
  const string ATLAS_2016_I1449082::measStr [] = {"incl","mtt","beta","ptt"}; 
  const string ATLAS_2016_I1449082::rangeStr[][kNbins] = { { "0_1", "1_2"}, //inclusive (dummy)
							   { "0_500", "500_2000"}, //mtt [GeV]
							   { "0_0.6", "0.6_1.0"}, //beta
							   { "0_30" , "30_1000"} }; //pttt

  const double ATLAS_2016_I1449082::MW = 80.300*GeV; 
  const double ATLAS_2016_I1449082::Mtop = 172.5*GeV; 

  
  // The hook for the plugin system
  DECLARE_RIVET_PLUGIN(ATLAS_2016_I1449082);
}
